#    Prana RC
#    Copyright (C) 2020 Dmitry Berezovsky
#
#    prana is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    prana is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

import argparse
import asyncio
import json
import sys
import traceback
from asyncio import AbstractEventLoop, CancelledError
from enum import Enum
from typing import Optional

from prana_rc.entity import Speed, PranaState
from prana_rc.service import PranaDeviceManager


class OutputFormat(Enum):
    TEXT = "text"
    JSON = "json"

    def __str__(self):
        return self.value


class CLI:
    verbose_mode = False

    @classmethod
    def print_data(cls, msg: str):
        print(msg)

    @classmethod
    def print_info(cls, msg: str):
        print(msg, file=sys.stderr)

    @classmethod
    def print_error(cls, exception):
        print(" -> ERROR: " + str(exception), file=sys.stderr)
        if cls.verbose_mode:
            print("--------------")
            traceback.print_exc(file=sys.stderr)

    @classmethod
    def print_debug(cls, string_to_print):
        if cls.verbose_mode:
            print("[DEBUG] " + string_to_print, file=sys.stderr)

    @classmethod
    def print_state(cls, state: PranaState, output_format: OutputFormat = OutputFormat.TEXT):
        if output_format == OutputFormat.TEXT:
            cls.print_data(str(state))
        elif output_format == OutputFormat.JSON:
            state_dict = state.to_dict()
            state_dict["timestamp"] = None  # Todo: Implement proper datetime converter
            cls.print_data(json.dumps(state_dict))

    @classmethod
    def print_version(cls, version_obj: dict, output_format: OutputFormat = OutputFormat.TEXT):
        if output_format == OutputFormat.TEXT:
            cls.print_data(str(version_obj.get("version")))
        elif output_format == OutputFormat.JSON:
            cls.print_data(json.dumps(version_obj))


class CliExtension(object):
    """
    Allows to extend CLI interface
    """

    COMMAND_NAME: Optional[str] = None
    COMMAND_DESCRIPTION: Optional[str] = None

    def __init__(
        self,
        parser: argparse.ArgumentParser,
        device_manager: PranaDeviceManager,
        loop: AbstractEventLoop,
    ):
        super().__init__()
        self.parser = parser
        self.__devce_manager = device_manager
        self.__loop = loop

    @staticmethod
    def _ensure_device_arg(args):
        if args.device is None:
            raise ValueError("Device is not set. Please check --device option.")

    async def connect_to_device(self, args):
        self._ensure_device_arg(args)
        CLI.print_info("Connecting to {}...".format(args.device))
        attempts_left = 10
        attempt_number = 1
        while attempts_left > 0:
            try:
                device = await self.device_manager.connect(args.device, args.timeout, attempts=1)
                CLI.print_info("   Connected")
                return device
            except CancelledError as e:
                CLI.print_info("Connect to device routine interrupted")
                raise e
            except Exception as e:
                attempts_left -= 1
                attempt_number += 1
                CLI.print_error(e)
                CLI.print_info("Reconnecting... Attempt #{}".format(attempt_number))
                await asyncio.sleep(1)
        raise RuntimeError("Unable to connect after {} attempts".format(attempt_number - 1))

    @property
    def device_manager(self) -> PranaDeviceManager:
        return self.__devce_manager

    @property
    def loop(self):
        return self.__loop

    @classmethod
    def setup_parser(cls, parser: argparse.ArgumentParser):
        pass

    def handle(self, args):
        raise NotImplementedError()


def register_global_arguments(parser: argparse.ArgumentParser):
    parser.add_argument(
        "-v",
        "--verbose",
        dest="verbose",
        action="store_true",
        required=False,
        help="If set more verbose output will used",
        default=False,
    )
    parser.add_argument(
        "-i",
        "--iface",
        dest="iface",
        action="store",
        required=False,
        type=str,
        help="Bluetooth interface to be used",
        default="hci0",
    )
    parser.add_argument(
        "-d",
        "--device",
        dest="device",
        action="store",
        required=False,
        type=str,
        help="Mac address of the prana device to connect to device. Required for the most of commands",
    )
    parser.add_argument(
        "-t",
        "--timeout",
        dest="timeout",
        action="store",
        required=False,
        type=int,
        help="Time in seconds to wait for device",
        default=3,
    )
    parser.add_argument(
        "-f",
        "--format",
        dest="format",
        action="store",
        required=False,
        type=OutputFormat,
        choices=list(OutputFormat),
        default=OutputFormat.TEXT,
        help="Output format, e.g. text, json, etc.",
    )


def parse_bool_val(v: str) -> bool:
    if isinstance(v, bool):
        return v
    if v.lower() in ("on", "yes", "true", "t", "y", "1"):
        return True
    elif v.lower() in ("off", "no", "false", "f", "n", "0"):
        return False
    else:
        raise argparse.ArgumentTypeError("Boolean value expected.")


def parse_speed_str(v: str) -> Speed:
    return Speed.from_str(v)


def parse_brightness_value(v: str) -> int:
    v_int: int = int(v) if not isinstance(v, int) else v
    if v_int < 1 or v_int > 6:
        raise argparse.ArgumentTypeError("Brightness value should be between 1 and 6.")
    return v_int



#    Prana RC
#    Copyright (C) 2020 Dmitry Berezovsky
#
#    prana is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    prana is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

import abc
import json

from sizzlews.client.common import SizzleWsAsyncClient
from typing import List

from prana_rc import utils
from prana_rc.contrib.api import (
    PranaRCAsyncFacade,
    SetStateDTO,
    DEFAULT_TIMEOUT,
    DEFAULT_ATTEMPTS,
    PranaStateDTO,
    PranaDeviceInfoDTO,
)
from prana_rc.contrib.api.dto import PranaHealthCheckResultDTO


class PranaRCAsyncClient(SizzleWsAsyncClient, PranaRCAsyncFacade, metaclass=abc.ABCMeta):
    async def discover(self, timeout=DEFAULT_TIMEOUT) -> List[PranaDeviceInfoDTO]:
        # TODO: set expected response time. Requires ws-sizzle update to support lists
        return utils.safe_cast(
            List[PranaDeviceInfoDTO], await self.async_invoke("prana.discover", timeout, expected_response_type=None)
        )

    async def get_state(self, address: str, timeout=DEFAULT_TIMEOUT, attempts=DEFAULT_ATTEMPTS) -> PranaStateDTO:
        return utils.safe_cast(
            PranaStateDTO,
            await self.async_invoke(
                "prana.get_state", address, timeout, attempts, expected_response_type=PranaStateDTO
            ),
        )

    async def set_state(
        self,
        address: str,
        state: SetStateDTO,
        timeout=DEFAULT_TIMEOUT,
        attempts=DEFAULT_ATTEMPTS,
    ) -> PranaStateDTO:
        return utils.safe_cast(
            PranaStateDTO,
            await self.async_invoke(
                "prana.set_state", address, json.loads(state.json()), timeout, expected_response_type=PranaStateDTO
            ),
        )

    async def healthcheck(self) -> PranaHealthCheckResultDTO:
        return utils.safe_cast(
            PranaHealthCheckResultDTO,
            await self.async_invoke("prana.healthcheck", expected_response_type=PranaHealthCheckResultDTO),
        )

    async def disconnect(self) -> None:
        await self.async_invoke("prana.disconnect_all")



#    Prana RC
#    Copyright (C) 2020 Dmitry Berezovsky
#
#    prana is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    prana is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

import asyncio
import datetime
import logging
import struct
from asyncio import AbstractEventLoop, Lock
from math import log2
from typing import Dict, List, Union, Optional

import bleak
from bleak.exc import BleakDBusError

from prana_rc import utils
from prana_rc.entity import PranaState, PranaDeviceInfo, Speed, PranaSensorsState
from prana_rc.utils import none_throws


class PranaDeviceManager(object):
    PRANA_DEVICE_NAME_PREFIXES = ["PRNAQaq", "PRANA"]

    def __init__(self, iface: str = "hci0", loop: Optional[AbstractEventLoop] = None) -> None:
        self.__ble_interface = iface
        self.__loop = loop
        self.__logger = logging.getLogger(self.__class__.__name__)
        self.__managed_devices: Dict["str", PranaDevice] = {}
        self.__lock = Lock()

    @classmethod
    def __is_prana_device(cls, dev: "bleak.backends.device.BLEDevice"):
        return dev.name and len(list(filter(dev.name.startswith, cls.PRANA_DEVICE_NAME_PREFIXES))) > 0

    @classmethod
    def __prana_dev_name_2_name(cls, dev_name: str):
        name = dev_name
        if dev_name:
            for prefix in cls.PRANA_DEVICE_NAME_PREFIXES:
                if dev_name.startswith(prefix):
                    name = dev_name.replace(prefix, "", 1)
                    break
        return name.strip()

    @classmethod
    def __addr_for_target(cls, target: Union[str, PranaDeviceInfo]) -> str:
        if isinstance(target, PranaDeviceInfo):
            address = target.address
        elif isinstance(target, str):
            address = target
        else:
            raise ValueError("Device must be specified either by mac address or by PranaDeviceInfo instance")
        return address

    async def discover(self, timeout: int = 5) -> List[PranaDeviceInfo]:
        """
        Listens to devices advertisement for TIMEOUT seconds and returns the list of discovered devices
        :param timeout: time to wait for devices in seconds
        :return: list of discovered devices
        """
        async with self.__lock:
            discovered_devs = await bleak.discover(timeout)
        return list(
            map(
                lambda dev: PranaDeviceInfo(
                    address=dev.address,
                    bt_device_name=(dev.name or "").strip(),
                    name=self.__prana_dev_name_2_name(dev.name),
                    rssi=dev.rssi,
                ),
                filter(PranaDeviceManager.__is_prana_device, discovered_devs),
            )
        )

    async def connect(self, target: Union[str, PranaDeviceInfo], timeout: float = 5, attempts=1) -> "PranaDevice":
        address = self.__addr_for_target(target)
        device = self.__managed_devices.get(address, None)
        if device is None:  # If not found in managed devices list
            device = PranaDevice(address, self.__loop, self.__ble_interface)
            self.__managed_devices[address] = device
        # if not await device.is_connected():
        attempts_left = attempts
        while attempts_left > 0:
            current_attempt = (attempts - attempts_left) + 1
            try:
                await device.connect(timeout)
                return device
            except Exception as e:
                if attempts == 1:
                    raise e
                self.__logger.warning("Connection failed. Attempt #{} Re-connecting...".format(current_attempt))
                attempts_left -= 1
                await asyncio.sleep(0.5)
        raise RuntimeError("Connection to device {} failed after {} attempts".format(address, attempts))

    async def disconnect_all(self):
        for addr, dev in list(self.__managed_devices.items()):
            try:
                self.__logger.info("Disconnecting {}...".format(dev.address))
                await dev.disconnect()
                del self.__managed_devices[addr]
            except Exception as e:
                self.__logger.exception("Unable to disconnect device {}: {}".format(dev.address, str(e)))

    def get_connected_devices_addresses(self) -> List[str]:
        """
        Returns a list of MAC addresses of the currently connected devices.
        Note: this method won't verify connection status so it might return 'dead' connections as well.
        """
        return list(self.__managed_devices.keys())


class PranaDevice(object):
    CONTROL_SERVICE_UUID = "0000baba-0000-1000-8000-00805f9b34fb"
    CONTROL_RW_CHARACTERISTIC_UUID = "0000cccc-0000-1000-8000-00805f9b34fb"
    STATE_MSG_PREFIX = b"\xbe\xef"
    MAX_BRIGHTNESS = 6

    class Cmd:
        ENABLE_HIGH_SPEED = bytearray([0xBE, 0xEF, 0x04, 0x07])
        ENABLE_NIGHT_MODE = bytearray([0xBE, 0xEF, 0x04, 0x06])
        TOGGLE_FLOW_LOCK = bytearray([0xBE, 0xEF, 0x04, 0x09])
        TOGGLE_HEATING = bytearray([0xBE, 0xEF, 0x04, 0x05])
        TOGGLE_WINTER_MODE = bytearray([0xBE, 0xEF, 0x04, 0x16])

        SPEED_UP = bytearray([0xBE, 0xEF, 0x04, 0x0C])
        SPEED_DOWN = bytearray([0xBE, 0xEF, 0x04, 0x0B])
        SPEED_IN_UP = bytearray([0xBE, 0xEF, 0x04, 0x0E])
        SPEED_IN_DOWN = bytearray([0xBE, 0xEF, 0x04, 0x0F])
        SPEED_OUT_UP = bytearray([0xBE, 0xEF, 0x04, 0x11])
        SPEED_OUT_DOWN = bytearray([0xBE, 0xEF, 0x04, 0x12])

        FLOW_IN_OFF = bytearray([0xBE, 0xEF, 0x04, 0x0D])
        FLOW_OUT_OFF = bytearray([0xBE, 0xEF, 0x04, 0x10])

        STOP = bytearray([0xBE, 0xEF, 0x04, 0x01])
        READ_STATE = bytearray([0xBE, 0xEF, 0x05, 0x01, 0x00, 0x00, 0x00, 0x00, 0x5A])
        READ_DEVICE_DETAILS = bytearray([0xBE, 0xEF, 0x05, 0x02, 0x00, 0x00, 0x00, 0x00, 0x5A])
        CHANGE_BRIGHTNESS = bytearray([0xBE, 0xEF, 0x04, 0x02])

    def __init__(
        self,
        target: Union[str, PranaDeviceInfo],
        loop: Optional[AbstractEventLoop] = None,
        iface: str = "hci0",
    ) -> None:
        self.__address = None
        if isinstance(target, PranaDeviceInfo):
            self.__address = target.address
        elif isinstance(target, str):
            self.__address = target
        else:
            raise ValueError(
                "PranaDevice constructor error: Target must be either mac address or PranaDeviceInfo instance"
            )
        self.__iface = iface
        self.__client = self.__new_client()
        self.__has_connect_attempts = False
        self.__notification_bytes: Optional[bytearray] = None
        self.__state: Optional[PranaState] = None
        self.__read_state_event: Optional[asyncio.Event] = None
        self.__lock = Lock()
        self.__logger = logging.getLogger(self.__class__.__name__)

    def __new_client(self) -> bleak.BleakClient:
        return bleak.BleakClient(self.__address, device=self.__iface)

    async def __verify_connected(self):
        if not await self.is_connected():
            raise RuntimeError("Illegal state: device must be connected before running any commands")

    def notification_handler(self, sender, data):
        """Simple notification handler which prints the data received."""
        self.__notification_bytes = data
        # Notify waiters
        if self.__read_state_event is not None:
            self.__read_state_event.set()

    async def connect(self, timeout: float = 2):
        async with self.__lock:
            if not await self.is_connected():
                try:
                    await self.__client.connect(timeout=timeout)
                    self.__has_connect_attempts = True
                except BleakDBusError:
                    await self.__client.disconnect()
                    self.__client = self.__client = self.__new_client()
                    raise
                await self.__client.start_notify(self.CONTROL_RW_CHARACTERISTIC_UUID, self.notification_handler)
                # TODO: shall we subscribe for disconnect callback and change status?
                # TODO: Verify prana service exists to ensure it is prana device

    async def disconnect(self):
        async with self.__lock:
            await self.__client.disconnect()

    async def is_connected(self):
        if not self.__has_connect_attempts:
            return False
        try:
            return self.__client.is_connected
        except Exception:
            self.__logger.error("Is Connected: Failed to verify connection status")
            return False

    async def _send_command(self, command: bytearray, expect_reply=False):
        async with self.__lock:
            # Invalidate state
            self.__state = None
            await self.__client.write_gatt_char(self.CONTROL_RW_CHARACTERISTIC_UUID, command, response=expect_reply)
            if expect_reply:
                self.__read_state_event = asyncio.Event()
                await asyncio.wait_for(self.__wait_for_read_event(), timeout=1)
                return self.__notification_bytes
            # await asyncio.sleep(0.6)
            # result = await self.__client.read_gatt_char(self.CONTROL_RW_CHARACTERISTIC_UUID, use_cached=False)
            # if expect_reply:
            #     return result

    async def set_high_speed(self):
        await self.__verify_connected()
        await self._send_command(self.Cmd.ENABLE_HIGH_SPEED)

    async def speed_up(self):
        await self.__verify_connected()
        await self._send_command(self.Cmd.SPEED_UP)

    async def speed_down(self):
        await self.__verify_connected()
        await self._send_command(self.Cmd.SPEED_DOWN)

    async def set_low_speed(self):
        await self.__verify_connected()
        await self._send_command(self.Cmd.ENABLE_NIGHT_MODE)

    async def set_night_mode(self):
        await self.__verify_connected()
        await self._send_command(self.Cmd.ENABLE_NIGHT_MODE)

    async def set_normal_speed(self):
        await self.set_speed(Speed.SPEED_3)

    async def set_speed(self, speed: Speed):
        if speed == Speed.LOW:
            await self.set_low_speed()
        elif speed == Speed.HIGH:
            await self.set_high_speed()
        elif speed == Speed.OFF:
            await self.turn_off()
        else:
            # Other speeds couldn't be set with one single command
            direction_up = speed.value <= 5
            if direction_up:
                await self.set_low_speed()
                counter = Speed.LOW.value
                while counter != speed.value:
                    await self.speed_up()
                    counter += 1
            else:
                await self.set_high_speed()
                counter = Speed.HIGH.value
                while counter != speed.value:
                    await self.speed_down()
                    counter -= 1

    async def set_brightness(self, brightness: int):
        if brightness < 0 or brightness > 6:
            raise ValueError("brightness value must be in range 0-6")
        original_state = await self.read_state()
        original_brightness = none_throws(original_state.brightness)
        if brightness == original_brightness:
            return
        if brightness > original_brightness:
            counter = brightness - original_brightness
        else:
            counter = brightness + (self.MAX_BRIGHTNESS - original_brightness)
        while counter > 0:
            await self.brightness_up()
            counter -= 1

    async def set_brightness_pct(self, brightness_pct: int):
        """
        Set brightness in percents (0-100)
        :param brightness_pct: integer in 0-100 range
        :return:
        """
        if brightness_pct < 0 or brightness_pct > 100:
            raise ValueError("brightness_pct is percent value (range 0-100)")
        return await self.set_brightness(round(self.MAX_BRIGHTNESS * brightness_pct / 100))

    async def brightness_up(self):
        await self.__verify_connected()
        await self._send_command(self.Cmd.CHANGE_BRIGHTNESS)

    async def set_heating(self, enable: bool):
        state = await self.read_state()
        if state.mini_heating_enabled != enable:
            await self._send_command(self.Cmd.TOGGLE_HEATING)

    async def set_winter_mode(self, enable: bool):
        state = await self.read_state()
        if state.winter_mode_enabled != enable:
            await self._send_command(self.Cmd.TOGGLE_WINTER_MODE)

    async def turn_off(self):
        await self.__verify_connected()
        await self._send_command(self.Cmd.STOP)

    async def turn_on(self, speed=Speed.SPEED_3):
        await self.set_speed(speed)

    def __parse_state(self, data: bytearray) -> Optional[PranaState]:
        if not data[:2] == self.STATE_MSG_PREFIX:
            return None
        self.__logger.debug("State data:")
        self.__logger.debug(",".join("0x{:02x}".format(x) for x in data))
        s = PranaState()
        s.timestamp = datetime.datetime.now()
        s.brightness = int(log2(data[12]) + 1)
        s.speed_locked = int(data[26] / 10)
        s.speed_in = int(data[30] / 10)
        s.speed_out = int(data[34] / 10)
        s.auto_mode = bool(data[20])
        s.night_mode = bool(data[16])
        s.flows_locked = bool(data[22])
        s.is_on = bool(data[10])
        s.mini_heating_enabled = bool(data[14])
        s.winter_mode_enabled = bool(data[42])
        s.is_input_fan_on = bool(data[28])
        s.is_output_fan_on = bool(data[32])
        # Reading sensors
        sensors = PranaSensorsState()
        sensors.humidity = int(data[60] - 128)
        sensors.pressure = 512 + int(data[78])
        # co2 and voc
        sensors.co2 = int(struct.unpack_from(">h", data, 61)[0] & 0b0011111111111111)
        sensors.voc = int(struct.unpack_from(">h", data, 63)[0] & 0b0011111111111111)
        if 0 < sensors.co2 < 10000:
            # Different version of firmware ???
            sensors.temperature_in = float(struct.unpack_from(">h", data, 51)[0] & 0b0011111111111111) / 10.0
            sensors.temperature_out = float(struct.unpack_from(">h", data, 54)[0] & 0b0011111111111111) / 10.0
        else:
            sensors.temperature_in = float(data[49]) / 10
            sensors.temperature_out = float(data[55]) / 10
        # Add sensors to the state only in case device has corresponding hardware
        if sensors.humidity > 0:
            s.sensors = sensors
        return s

    async def read_state(self, force_read: bool = False) -> PranaState:
        """
        Read state from the devcie and return it as an object
        :param force_read: If set, cached state will be ignored and read command to the device will be generated
        :return:
        """
        await self.__verify_connected()
        state_bin = await self._send_command(self.Cmd.READ_STATE, expect_reply=True)
        state = self.__parse_state(state_bin)
        if state is not None:
            self.__state = state
        return utils.none_throws(state)

    async def __wait_for_read_event(self):
        if self.__read_state_event is not None:
            await self.__read_state_event.wait()

    def __has_relevant_state(self) -> bool:
        return not (
            self.__state is None
            or (datetime.datetime.now() - utils.none_throws(self.__state.timestamp)).total_seconds() > 60
        )

    @property
    def address(self):
        return self.__address



#    Prana RC
#    Copyright (C) 2020 Dmitry Berezovsky
#
#    prana is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    prana is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

import datetime
from asyncio import Lock
from asyncio.events import AbstractEventLoop

from jsonrpc import Dispatcher
from pydantic import validate_arguments
from sizzlews.server.annotation import rpc_method
from sizzlews.server.common import MethodDiscoveryMixin, SizzleWSHandler
from typing import List, Optional

from prana_rc import utils, __version__
from prana_rc.contrib.api import (
    PranaRCAsyncFacade,
    DEFAULT_TIMEOUT,
    DEFAULT_ATTEMPTS,
    SetStateDTO,
    PranaStateDTO,
    PranaDeviceInfoDTO,
)
from prana_rc.contrib.api.dto import PranaHealthCheckResultDTO
from prana_rc.entity import Mode, PranaDeviceInfo, PranaState
from prana_rc.service import PranaDeviceManager, PranaDevice


class ToDTO(object):
    @classmethod
    def prana_device_info(cls, obj: Optional[PranaDeviceInfo]) -> Optional[PranaDeviceInfoDTO]:
        if obj is None:
            return None
        return PranaDeviceInfoDTO(
            address=obj.address,
            bt_device_name=obj.bt_device_name,
            name=obj.name,
            rssi=obj.rssi,
        )

    @classmethod
    def prana_state(cls, obj: Optional[PranaState]) -> Optional[PranaStateDTO]:
        if obj is None:
            return None
        return PranaStateDTO.parse_obj(obj.to_dict())


class PranaRCApiHandler(MethodDiscoveryMixin, SizzleWSHandler, PranaRCAsyncFacade):
    METHOD_PREFXIX = "prana."

    def __init__(
        self,
        device_manager: PranaDeviceManager,
        loop: AbstractEventLoop,
        dispatcher: Dispatcher = None,
        expose_version_api=True,
        expose_ping_api=True,
    ) -> None:
        super().__init__(dispatcher, expose_version_api, expose_ping_api)
        self.__device_manager = device_manager
        # self.__devices_pool = {}  # type: Dict[str, PranaDevice]
        self.__loop = loop
        self.__lock = Lock()

    async def get_connected_prana_device(
        self, device_addr: str, timeout=DEFAULT_TIMEOUT, attempts=DEFAULT_ATTEMPTS
    ) -> PranaDevice:
        async with self.__lock:
            prana_device = await self.__device_manager.connect(device_addr, timeout, attempts)
            return prana_device

    @rpc_method
    async def discover(self, timeout=4) -> List[PranaDeviceInfoDTO]:
        res = await self.__device_manager.discover(timeout)
        return [utils.none_throws(ToDTO.prana_device_info(d)) for d in res]

    @rpc_method
    async def get_state(self, address: str, timeout=DEFAULT_TIMEOUT, attempts=DEFAULT_ATTEMPTS) -> PranaStateDTO:
        prana_device = await self.get_connected_prana_device(address, timeout, attempts)
        state = await prana_device.read_state()
        return utils.none_throws(ToDTO.prana_state(state))

    @rpc_method
    @validate_arguments
    async def set_state(
        self,
        address: str,
        state: SetStateDTO,
        timeout=DEFAULT_TIMEOUT,
        attempts=DEFAULT_ATTEMPTS,
    ) -> PranaStateDTO:
        features = [state.speed, state.mode, state.winter_mode, state.heating, state.brightness, state.brightness_pct]
        if all(v is None for v in features):
            raise ValueError("At least one parameter must be set. Check your arguments.")
        prana_device = await self.get_connected_prana_device(address, timeout, attempts)
        if state.speed is not None:
            await prana_device.set_speed(state.speed)
        if state.heating is not None:
            await prana_device.set_heating(state.heating)
        if state.winter_mode is not None:
            await prana_device.set_winter_mode(state.winter_mode)
        if state.mode is not None:
            if state.mode == Mode.NIGHT:
                await prana_device.set_night_mode()
            elif state.mode == Mode.NORMAL:
                await prana_device.set_normal_speed()
            elif state.mode == Mode.HIGH:
                await prana_device.set_high_speed()
        if state.brightness is not None or state.brightness_pct is not None:
            if state.brightness is not None and state.brightness_pct is not None:
                raise ValueError('Use either "brightness" or "brightness_pct" but not both at the same time.')
            if state.brightness is not None:
                await prana_device.set_brightness(state.brightness)
            if state.brightness_pct is not None:
                await prana_device.set_brightness_pct(state.brightness_pct)
        new_state = await prana_device.read_state()
        return utils.none_throws(ToDTO.prana_state(new_state))

    @rpc_method
    async def healthcheck(self) -> PranaHealthCheckResultDTO:
        return PranaHealthCheckResultDTO(
            version=__version__.__version__,
            timestamp=datetime.datetime.now(),
            current_connections=self.__device_manager.get_connected_devices_addresses(),
        )

    @rpc_method
    async def disconnect_all(self) -> None:
        await self.__device_manager.disconnect_all()



#    Prana RC
#    Copyright (C) 2020 Dmitry Berezovsky
#
#    prana is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    prana is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

from typing import Optional, TypeVar, Type, Any

__all__ = ("none_throws", "safe_cast")


_T = TypeVar("_T")


def none_throws(optional: Optional[_T], message: str = "Unexpected `None`") -> _T:
    """Convert an optional to its value. Raises an `AssertionError` if the
    value is `None`"""
    if optional is None:
        raise AssertionError(message)
    return optional


def safe_cast(new_type: Type[_T], value: Any) -> _T:
    """safe_cast will change the type checker's inference of x if it was
    already a subtype of what we are casting to, and error otherwise."""
    return value  # type: ignore[no-any-return]


# ===================================



#    Prana RC
#    Copyright (C) 2020 Dmitry Berezovsky
#
#    prana is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    prana is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

import abc

from typing import List

from prana_rc.contrib.api.dto import SetStateDTO, PranaDeviceInfoDTO, PranaStateDTO

DEFAULT_TIMEOUT = 5
DEFAULT_ATTEMPTS = 10


class PranaRCAsyncFacade(abc.ABC):
    @abc.abstractmethod
    async def discover(self, timeout=DEFAULT_TIMEOUT) -> List[PranaDeviceInfoDTO]:
        pass

    @abc.abstractmethod
    async def get_state(self, address: str, timeout=DEFAULT_TIMEOUT, attempts=DEFAULT_ATTEMPTS) -> PranaStateDTO:
        pass

    @abc.abstractmethod
    async def set_state(
        self,
        address: str,
        state: SetStateDTO,
        timeout=DEFAULT_TIMEOUT,
        attempts=DEFAULT_ATTEMPTS,
    ) -> PranaStateDTO:
        pass



#    Prana RC
#    Copyright (C) 2020 Dmitry Berezovsky
#
#    prana is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    prana is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

import datetime
from enum import Enum

from typing import NamedTuple, List, Optional


class Speed(Enum):
    OFF = 0
    LOW = 1
    HIGH = 10
    SPEED_2 = 2
    SPEED_3 = 3
    SPEED_4 = 4
    SPEED_5 = 5
    SPEED_6 = 6
    SPEED_7 = 7
    SPEED_8 = 8
    SPEED_9 = 9

    @classmethod
    def all_options(cls) -> List[str]:
        return ["low", "l", "high", "h", "off", "stop", "2", "3", "4", "5", "6", "7", "8", "9"]

    @classmethod
    def from_str(cls, speed: str) -> "Speed":
        speed = str(speed).lower().strip()
        if speed in ["low", "l"]:
            return cls.LOW
        if speed in ["high", "h"]:
            return cls.HIGH
        if speed in ["off", "stop"]:
            return cls.OFF
        try:
            speed_int = int(speed)
            if 0 <= speed_int <= 10:
                return cls(speed_int)
        except ValueError:
            pass
        raise ValueError("String {} is not valid speed identifier".format(speed))

    def to_int(self) -> int:
        return int(self.value)


class Mode(Enum):
    NORMAL = "normal"
    NIGHT = "night"
    HIGH = "high"


class PranaDeviceInfo(NamedTuple):
    address: str
    bt_device_name: str
    name: str
    rssi: int


class PranaSensorsState(object):
    def __init__(self) -> None:
        self.temperature_in: Optional[float] = None
        self.temperature_out: Optional[float] = None
        self.humidity: Optional[int] = None
        self.pressure: Optional[int] = None
        self.voc: Optional[int] = None
        self.co2: Optional[int] = None

    def __repr__(self):
        return (
            "Temperature: (in: {}, out: {}), Humidity: {}, Pressure: {}".format(
                self.temperature_in, self.temperature_out, self.humidity, self.pressure
            )
            + ", VOC: {}, CO2: {}".format(self.voc, self.co2)
            if self.co2 is not None or self.voc is not None
            else ""
        )

    def to_dict(self) -> dict:
        return dict(
            temperature_in=self.temperature_in,
            temperature_out=self.temperature_out,
            humidity=self.humidity,
            pressure=self.pressure,
            voc=self.voc,
            co2=self.co2,
        )


class PranaState(object):
    def __init__(self) -> None:
        self.speed_locked: Optional[int] = None
        self.speed_in: Optional[int] = None
        self.speed_out: Optional[int] = None
        self.night_mode: Optional[bool] = None
        self.auto_mode: Optional[bool] = None
        self.flows_locked: Optional[bool] = None
        self.is_on: Optional[bool] = None
        self.mini_heating_enabled: Optional[bool] = None
        self.winter_mode_enabled: Optional[bool] = None
        self.is_input_fan_on: Optional[bool] = None
        self.is_output_fan_on: Optional[bool] = None
        self.brightness: Optional[int] = None
        self.sensors: Optional[PranaSensorsState] = None
        self.timestamp: Optional[datetime.datetime] = None

    @property
    def speed(self):
        if not self.is_on:
            return 0
        return self.speed_locked if self.flows_locked else int((self.speed_in + self.speed_out) / 2)

    def __repr__(self):
        res = "Prana state: {}, Speed: {}, Winter Mode: {}, Heating: {}, Flows locked: {}, Brightness: {}".format(
            "RUNNING" if self.is_on else "IDLE",
            self.speed,
            self.winter_mode_enabled,
            self.mini_heating_enabled,
            self.flows_locked,
            self.brightness,
        )
        if self.sensors is not None:
            res += " Sensors: {" + repr(self.sensors) + "}"
        return res

    def to_dict(self) -> dict:
        return dict(
            speed_locked=self.speed_locked,
            speed_in=self.speed_in,
            speed_out=self.speed_out,
            night_mode=self.night_mode,
            auto_mode=self.auto_mode,
            flows_locked=self.flows_locked,
            is_on=self.is_on,
            mini_heating_enabled=self.mini_heating_enabled,
            winter_mode_enabled=self.winter_mode_enabled,
            is_input_fan_on=self.is_input_fan_on,
            is_output_fan_on=self.is_output_fan_on,
            timestamp=self.timestamp if self.timestamp is not None else None,
            speed=self.speed,
            brightness=self.brightness,
            sensors=self.sensors.to_dict() if self.sensors is not None else None,
        )


# TODO: Deprecated???
class ToApiDict(object):
    @classmethod
    def prana_device_info(cls, obj: Optional[PranaDeviceInfo]) -> Optional[dict]:
        if obj is None:
            return None
        return dict(
            address=obj.address,
            bt_device_name=obj.bt_device_name,
            name=obj.name,
            rssi=obj.rssi,
        )

    @classmethod
    def prana_state(cls, obj: Optional[PranaState]) -> Optional[dict]:
        if obj is None:
            return None
        return obj.to_dict()



#    Prana RC
#    Copyright (C) 2020 Dmitry Berezovsky
#
#    prana is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    prana is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

import argparse
import asyncio
import importlib
import inspect
import pkgutil
import signal
from asyncio import CancelledError

from typing import Type, List

import prana_rc.__version__
from prana_rc.cli_utils import (
    CliExtension,
    register_global_arguments,
    CLI,
    parse_bool_val,
    parse_speed_str,
    parse_brightness_value,
)
from prana_rc.entity import Mode
from prana_rc.service import PranaDeviceManager

PRANA_RC_VERSION = prana_rc.__version__.__version__
SHUTDOWN_SIGNALS = (signal.SIGHUP, signal.SIGTERM, signal.SIGINT)

supplementary_parser = argparse.ArgumentParser(add_help=False)
register_global_arguments(supplementary_parser)

parser = argparse.ArgumentParser(
    prog="prana",
    add_help=True,
    description="CLI interface to manage Prana recuperators via BLE",
    formatter_class=argparse.RawDescriptionHelpFormatter,
    epilog="""
Version: {version}
--------------------------------------------------------------
Prana RC CLI Copyright (C) 2020 Dmitry Berezovsky
This program comes with ABSOLUTELY NO WARRANTY;
This is free software, and you are welcome to redistribute it
under certain conditions;
--------------------------------------------------------------
""".format(
        version=PRANA_RC_VERSION
    ),
)
register_global_arguments(parser)
command_parser = parser.add_subparsers(
    title="command",
    dest="cmd",
    description='Use "<command> -h" to get information ' "about particular command",
)


# =========== AVAILABLE COMMANDS =====================


class DiscoveryCLIExtension(CliExtension):
    COMMAND_NAME = "discover"
    COMMAND_DESCRIPTION = "Run discovery process to find available Prana devices nearby"

    async def handle(self, args):
        CLI.print_info("Starting discovery (timeout {}s)...".format(args.timeout))
        devs = await self.device_manager.discover()
        if len(devs) == 0:
            CLI.print_info("No devices found")
        else:
            for dev in devs:
                CLI.print_info(
                    "{} [{}] (identity: {}, rssi: {})".format(dev.name, dev.address, dev.bt_device_name, dev.rssi)
                )


class VersionCLIExtension(CliExtension):
    COMMAND_NAME = "version"
    COMMAND_DESCRIPTION = "Print version and exit"

    async def handle(self, args):
        version_obj = dict(version=PRANA_RC_VERSION)
        CLI.print_version(version_obj, args.format)


class ReadStateCLIExtension(CliExtension):
    COMMAND_NAME = "status"
    COMMAND_DESCRIPTION = "Read state of the device and print back to terminal"

    async def handle(self, args):
        device = await self.connect_to_device(args)
        state = await device.read_state()
        CLI.print_state(state, args.format)


class SetCLIExtension(CliExtension):
    COMMAND_NAME = "set"
    COMMAND_DESCRIPTION = "Set speed"

    @classmethod
    def setup_parser(cls, parser: argparse.ArgumentParser):
        parser.add_argument(
            "-s",
            "--speed",
            dest="speed",
            action="store",
            required=False,
            type=parse_speed_str,
            help="Speed to set. Could be one of: 0-10, off, high, low",
        )
        parser.add_argument(
            "-m",
            "--mode",
            dest="mode",
            action="store",
            required=False,
            type=Mode,
            help="Mode to set. Could be normal, night or high",
        )
        parser.add_argument(
            "-w",
            "--winter-mode",
            dest="winter_mode",
            action="store",
            required=False,
            type=parse_bool_val,
            help="Enable or disable winter mode (de-icing)",
        )
        parser.add_argument(
            "-q",
            "--heating",
            dest="heating",
            action="store",
            required=False,
            type=parse_bool_val,
            help="Enable or disable heating (mini-heating function in device manual)",
        )
        parser.add_argument(
            "-b",
            "--brightness",
            dest="brightness",
            action="store",
            required=False,
            type=parse_brightness_value,
            help="Set brightness (value between 1 and 6)",
        )

    async def handle(self, args: argparse.Namespace):
        features = [args.speed, args.mode, args.winter_mode, args.heating, args.brightness]
        if all(v is None for v in features):
            raise ValueError("At least one parameter must be set. Check your arguments.")
        device = await self.connect_to_device(args)
        if args.speed is not None:
            CLI.print_info("Setting speed to {}...".format(args.speed.value))
            await device.set_speed(args.speed)
        if args.heating is not None:
            CLI.print_info("Setting heating to {}...".format(args.heating))
            await device.set_heating(args.heating)
        if args.winter_mode is not None:
            CLI.print_info("Setting winter-mode to {}...".format(args.winter_mode))
            await device.set_winter_mode(args.winter_mode)
        if args.mode is not None:
            CLI.print_info("Setting mode to {}...".format(args.mode))
            if args.mode == Mode.NIGHT:
                await device.set_night_mode()
            elif args.mode == Mode.NORMAL:
                await device.set_normal_speed()
            elif args.mode == Mode.HIGH:
                await device.set_high_speed()
        if args.brightness is not None:
            CLI.print_info("Setting brightness to {}...".format(args.brightness))
            await device.set_brightness(args.brightness)

        # At the end let's print the new state
        state = await device.read_state()
        CLI.print_info("Recent device status:")
        CLI.print_state(state, args.format)


# =========== END OF COMMAND DEFINITIONS ==============


def read_global_args():
    known, unknown = supplementary_parser.parse_known_args()
    return known


def configure_subparser_for_cli_extension(
    ext: Type[CliExtension],
    parser: argparse.ArgumentParser,
    device_manager: PranaDeviceManager,
    loop: asyncio.AbstractEventLoop,
):
    ext.setup_parser(parser)
    ext_instance = ext(parser, device_manager, loop)
    parser.set_defaults(handler=ext_instance)


async def handle_wrapper(device_manager: PranaDeviceManager, args):
    try:
        await args.handler.handle(args)
    except CancelledError:
        pass
    except Exception as e:
        CLI.print_error(e)
    finally:
        await device_manager.disconnect_all()


async def on_shutdown(signal, loop, device_manager: PranaDeviceManager):
    CLI.print_error(f"Received exit signal {signal.name}...")
    tasks = [t for t in asyncio.all_tasks() if t is not asyncio.current_task()]

    [task.cancel() for task in tasks]
    await asyncio.gather(*tasks)
    await device_manager.disconnect_all()

    # loop.stop()


def discover_cli_extensions(target_package_name: str) -> List[Type]:
    extensions = []
    target_package = importlib.import_module(target_package_name)
    for loader, pkg_name, is_pkg in pkgutil.walk_packages(target_package.__path__):  # type: ignore
        # Check if we've got a valid extension package
        if is_pkg:
            full_name = ".".join((target_package.__name__, pkg_name))
            try:
                candidate_pkg = importlib.import_module(full_name)
                if hasattr(candidate_pkg, "is_available"):
                    if not candidate_pkg.is_available():  # type: ignore
                        continue  # Package is not supported
                    cli_module = importlib.import_module(".cli", full_name)
                    found_extensions = inspect.getmembers(
                        cli_module,
                        lambda member: inspect.isclass(member)
                        and member.__name__ != CliExtension.__name__
                        and issubclass(member, CliExtension),
                    )
                    extensions += [cls for name, cls in found_extensions]
            except ImportError:
                pass
    return extensions


def run_cli():
    # logging.basicConfig(level=logging.DEBUG)
    root_commands = [
        DiscoveryCLIExtension,
        SetCLIExtension,
        ReadStateCLIExtension,
        VersionCLIExtension,
    ]
    # Extra commands
    root_commands += discover_cli_extensions("prana_rc.contrib")
    global_args = read_global_args()
    CLI.verbose_mode = global_args.verbose

    loop = asyncio.get_event_loop()
    device_manager = PranaDeviceManager(iface=global_args.iface)

    # Process root commands
    for cmd in root_commands:
        cmd_subparser = command_parser.add_parser(cmd.COMMAND_NAME, help=cmd.COMMAND_DESCRIPTION)
        configure_subparser_for_cli_extension(cmd, cmd_subparser, device_manager, loop)

    args = parser.parse_args()
    if hasattr(args, "handler"):
        # Register on shutdown callback
        for s in SHUTDOWN_SIGNALS:
            loop.add_signal_handler(s, lambda s=s: asyncio.create_task(on_shutdown(s, loop, device_manager)))
        # Run main login
        try:
            loop.run_until_complete(asyncio.ensure_future(handle_wrapper(device_manager, args)))
        except KeyboardInterrupt:
            CLI.print_error("Process interrupted. Closing connections.")


if __name__ == "__main__":
    run_cli()



#    Prana RC
#    Copyright (C) 2020 Dmitry Berezovsky
#
#    prana is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    prana is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

import argparse
import asyncio
from asyncio import CancelledError

from sizzlews.server.tornado import (
    bootstrap_torando_rpc_application,
)

from prana_rc.cli_utils import CliExtension, CLI
from prana_rc.contrib.api.handler import PranaRCApiHandler
from prana_rc.service import PranaDeviceManager


class HttpServerCLIExtension(CliExtension):
    COMMAND_NAME = "http-server"
    COMMAND_DESCRIPTION = "Run HTTP server which exposes functionality via RPC interface"

    @classmethod
    def setup_parser(cls, parser: argparse.ArgumentParser):
        parser.add_argument(
            "-p",
            "--port",
            dest="http_port",
            action="store",
            required=False,
            type=int,
            default=8881,
            help="Port to bind http server.",
        )
        parser.add_argument(
            "-r",
            "--path",
            dest="http_path",
            action="store",
            required=False,
            type=str,
            default="/",
            help="Http path to bind rpc endpoint. E.g. /rpc will mount rpc server to "
            "http://localhost:<port>/rpc endpoint. If nothing is rpc will be mounted to the root. ",
        )

    async def handle(self, args: argparse.Namespace):
        CLI.print_info("Prana RC: Starting in HTTP server mode")
        device_manager = PranaDeviceManager(iface=args.iface)
        prana_api = PranaRCApiHandler(device_manager, asyncio.get_event_loop())
        bootstrap_torando_rpc_application(prana_api, args.http_port, args.http_path)
        CLI.print_info("HTTP: Listening on http://0.0.0.0:{}{}".format(args.http_port, args.http_path))
        try:
            while True:
                await asyncio.sleep(5)
        except CancelledError:
            CLI.print_info("Received shutdown signal. Closing connections...")
            await device_manager.disconnect_all()
            CLI.print_info("Connections closed")



#    Prana RC
#    Copyright (C) 2020 Dmitry Berezovsky
#
#    prana is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    prana is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

import datetime

import pydantic
from typing import Optional, List

from prana_rc.entity import Speed, Mode


class SetStateDTO(pydantic.BaseModel):
    speed: Optional[Speed] = None
    mode: Optional[Mode] = None
    winter_mode: Optional[bool] = None
    heating: Optional[bool] = None
    brightness: Optional[int] = None
    brightness_pct: Optional[int] = None


class PranaDeviceInfoDTO(pydantic.BaseModel):
    address: Optional[str] = None
    bt_device_name: Optional[str] = None
    name: Optional[str] = None
    rssi: Optional[int] = None


class PranaSensorsStateDTO(pydantic.BaseModel):
    temperature_in: Optional[float] = None
    temperature_out: Optional[float] = None
    humidity: Optional[int] = None
    pressure: Optional[int] = None
    voc: Optional[int] = None
    co2: Optional[int] = None


class PranaStateDTO(pydantic.BaseModel):
    speed_locked: Optional[int] = None
    speed_in: Optional[int] = None
    speed_out: Optional[int] = None
    brightness: Optional[int] = None
    night_mode: Optional[bool] = None
    auto_mode: Optional[bool] = None
    flows_locked: Optional[bool] = None
    is_on: Optional[bool] = None
    mini_heating_enabled: Optional[bool] = None
    winter_mode_enabled: Optional[bool] = None
    is_input_fan_on: Optional[bool] = None
    is_output_fan_on: Optional[bool] = None
    sensors: Optional[PranaSensorsStateDTO] = None
    timestamp: Optional[datetime.datetime] = None


class PranaHealthCheckResultDTO(pydantic.BaseModel):
    version: str
    timestamp: datetime.datetime
    current_connections: List[str]
